#!/usr/bin/env python3  
#
# Fietjes Extra Editor (FEE) Version 2023.02.18 #
# Autor   : DH0PE - DH0PE@ingolfpohl.de
# Licence : GPLv3
# Credits : MKA & son - hints for WIN10 Version
#         : Wolfgang - finding bugs and improvements  
#
# Code is starting at about line 240, so we have 240 lines of manual.
# »-------------------------------------------------------------------«
# What's it made for?
#
# It's a text editor, which can send parts of the text via hot-keys
# to a FORTH target.
# The answers coming from the target are sent to the terminal window
# from where the editor is started.
# Within the terminal window you can easily communicate with the
# target system.
# 
# The text can be split into blocks by putting in STOPPER lines.
# The block, where the cursor stands can be sent to the target
# via a hot-key (ctrl+return). This is made line by line and the
# editor watches out for "ok." to handle the target with maximum speed.
#
# If there is a leading \ in a line the transmission of line is
# skipped, because it's comment. 
# If there is an include-string ("include" or "\ include") at the
# first postion of a line, the editor takes the rest of the line
# as a path to a source code file, which is loaded and sent to
# the target.  
# 
# The edit function of the editor are derived from python's tkinter.
# I added some hot-keys for my convenience.
#
# Communication to the target can be done via serial interface or by
# telnet with for example an WiFi to serial bridge.
#
# »-------------------------------------------------------------------«
# What is needed to run it?
#
# * on Linux: Python 3 + tkinter + pyserial
# * on Win7:  python 3.8.6, pyserial, driver for USB-serial-adapter
# * on Win10: python 3.9.x, pyserial, driver for USB-serial-adapter
# * a serial interface, real or as USB-Adapter
# * a target board with Mecrisp Stellaris or Mecrisp Quintus
#
# »-------------------------------------------------------------------«
# How to start the program?
#
# * test for interfaces: python FEE.py
# * start command:       python FEE.py <interface> <baudrate>
# * or with telnet:      python FEE.py telnet <host-ip>:<port>
# * example on linux:    python FEE.py /dev/ttyUSB0 115200
# * example telnet:      python FEE.py telnet 192.168.4.1:23     
#
# »-------------------------------------------------------------------«
# Are there any useful command line parameters?
#
# There are some additional command line parameters, which can be
# added after the 2nd parameter:
#
#   DTRN   : inverts the DTR-Reset from active low (default) to high
#   RTSN   : inverts the RTS-Boot0 from active high (default) to low
#   RTSCTS : activates HW handshake via RTS/CTS
#   NOSAVE : switches off backup copy
#   FASTER : skip comments after a backslash on transfer to target
#   fn='font name' : replaces the default font 'mono' by 'font name'
#   fs=xx      : set font size, for example: fs=10 is default
#   fc=#rrggbb : set font color, default is #d5d2c8 (light grey) 
#   bc=#rrggbb : set background color, default is #343d46 (dark grey)
#   ac=#rrggbb : set active color, default is #b43104 (red)
#   cc=#rrggbb : set cursor color, default is #b43104 (red)
#   cw=xx      : set cursor width, default is 0 for block cursor
#   co=xxxx    : set cursor on time in ms, default is 1000
#   cf=xxxx    : set cursor off time in ms, default is 0
#   is='string': overrides the default include string '\ include'
#                default strings: 'include','#include','\ include'   
# caution: strings with blanks have to be put in ' '
# colors can be named according tkinter: #rrggbb or a color name
#
# »-------------------------------------------------------------------«
# Hotkeys
# 
# I just defined some hot keys, which only work within the editor
# window, not in the terminal window:
#
# menu "file":
# ctrl+o       = open source code file
# ctrl+r       = reload and overwrite actual text in editor
# ctrl+s       = save source code file
# ctrl+shift+s = write source code file as
# ctrl+shift+w = switch on/off logging target output
# ctrl+shift+q = quit the editor 
#
# menu "edit":
# ctrl+z       = undo
# ctrl+shift+z = redo
# ctrl+x       = cut
# ctrl+c       = copy
# ctrl+v       = paste
# ctrl+l       = cut line
# alt+l        = copy line
# ctrl+v       = paste line
# shift+enter  = try line - send to target
# ctrl+enter   = try block - send actual textblock to target
# alt+enter    = send block - send block to target and move block up
# ctrl+j       = insert a STOPPER line
# alt+j        = convert block - save textblock, replace it by "include <filename>"
# ctrl+shift+m = show/hide spaces
#
# menu "tools":
# ctrl-q       = clear the terminal window
# alt-q        = reset target  
#
# useful hotkeys without menu:
# strg+alt+m   = insert the magic character '»'
# strg+shift+t = mark actual stopper pair
#
# Hotkeys provided by Python's tkinter:
# ctrl+x = cut
# ctrl+v = paste
# ctrl+y = paste
# ctrl+c = copy
# ctrl+b = move cursor one character backward
# ctrl+n = move cursor one line forward
# ctrl+a = move cursor to first position in line
# ctrl+d = delete at cursor position
# ctrl+f = move cursor one character forward
# ctrl+h = backspace (delete)
# ctrl+p = move cursor one line backwards
# ctrl+i = insert a tabulator
# ctrl+t = move character right
# ctrl+e = place cursor at end of line
# ctrl+k = cut line right from cursor 
#
# »-------------------------------------------------------------------«
# Additional functions
#
# * empty lines and lines with \ in 1st column are not send to target
# * canceling rest of line after \ on request (FASTER)
# * auto backup save before critical operation into ./sav folder
# * auto backup if UART fails due power down or target lost
# * auto backup file names: YYYY-MM-DD_hh-mm-ss.sav
# * including external source files via "include path_to_sourcefile"
# * flash firmware for GD32VF103/STM32F103
# * supporting ESP-link via telnet with timeout avoidance
# * STOPPER overload by loading root file with magic word in 1st line
#   followed by the new stopper line in 2nd line
# * magic block replacement: If a block is ready to be outsourced to 
#   an include file, put »path_to_file in the first line of the block
#   and press alt+j. The block will be moved into the file 'path_to_file'
#   and will be replaced by: 'include path_to_file'. Without the magic
#   line a dialog window will ask for the filename.
# * inline baudrate switching " <code> \ »baud: <baudrate> "
#   where <code> switches baudrate on target and '\ »baud: <baudrate>'
#   switches baudrate off the editor.
# 
# »-------------------------------------------------------------------«
# How to install on Windows?
#
# 1) download python from https://python.org/downloads/windows
# 2) install python, use expert mode to set Windows path variables
# 3) include PIP
# 4) use PIP to install pyserial on command line: PIP install pyserial
# 5) install the driver for Your serial adapter (tested: Silabs CP2101)
# 6) press windows logo key + R and type cmd in the open box
# 7) on command line in terminal: go into folder where FEE is located
# 8) check FEE by entering: python FEE.py
# 9) FEE lists serial interfaces, choose one: python FEE.py COM23 115200
#
# »-------------------------------------------------------------------«
# How do I personally use the editor? 
# 
# I very often use it on linux with a cheap USB-to-serial adapter or
# with an WiFi-serial-bridge (like ESP-link).
#
# I keep my source code in a working root file and in additional
# source files to include. It looks like so:          
#
#   \ »--- start of text -------------------------------------------«
#   \ SOURCE CODE AREA 1 - comments and optional an "eraseflash"
#   \ Here I put normaly a lot of comments, headers etc.
#
#   eraseflash                    \ erase Flash except the kernel
#
#   \ »--- STOPPER -------------------------------------------------«
#   \ SOURCE CODE AREA 2 - well hung code, ready for Flash
#
#   compiletoflash                \ compile into Flash
#   include ../library/delete.txt  \ tool to delete last entry
#   include ../library/print.txt    \ some output for 4/8/16 bit
#
#   \ »--- STOPPER -------------------------------------------------«
#   \ SOURCE CODE AREA 3 - not yet ready for Flash but for RAM
#
#   forgetram                     \ epmty the RAM
#   compiletoram                   \ compile into RAM
#   include ../library/bit.txt      \ additional bit manipulations
#   include ../library/gpio.txt      \ GPIO registers and access
#
#   \ »--- STOPPER -------------------------------------------------«
#   \ SOURCE CODE AREA 4 - actual working area
#
#   : function_to_try  ( -- )     \ compiling in any case
#     1000 0 do                    \ repeat for maesurement 1000x
#       $0001 GPIOA ports!          \ set PA0
#       $0001 GPIOA portc!           \ reset PA0
#     loop                            \
#   ;                                  \
#
#   function_to_try                      \ call the compiled code 
#   .s                                    \ check the stack
#   del                                    \ delete last dictionary
#                                           \ entry for next trial
#
#   \ »--- STOPPER -------------------------------------------------«
#   \ SOURCE CODE AREA 5 and more - test code, draft code, etc
#   \ »--- end of text ---------------------------------------------«   
#
# I devide the source in different categories, each in a text block.
# If I have a new or screwed up system, I load the first 3 blocks to
# the target and get easy back to my working state.
# The "eraseflash" is in the 1st block because it goes through a reset,
# which needs a timeout. So You can't compile over it, at the moment.
# Areas 2 and 3 are more or less self explaining.
# It'll become interesting in area 4: here I test my code. I edit it
# mostly inside a word definition, compile it and run it by a hotkey.
# You see in the comments, that I use the del function from Matthias
# Koch, so the dictionary is not blown up while testing. When my code
# is ok, I compile the function only and move the text one block up.
# In areas greater then 4 I develop test code, alternative code or
# matching top-down code for the bottom-up code I write in area 4.
# When the project is finished I'll get source code in one block. All
# the code is trustful and ready for flash ;-)
#     
# »===================================================================«
# Sourcecode is starting exactly here :-)

# Default settings for editor -----------------------------------------
# may be overwritten by command line arguments or inline instructions

KopfZeile = "Fietjes Extra Editor (FEE) Version 2023-02-18 by DH0PE"

STOPPER_DEF = "\ »--- STOPPER -------------------------------------------------«\n"
STOPPER = STOPPER_DEF
stonu = 255  # number of characters to highlight on test purposes

fontname    = 'mono' # 'Bitstream Vera Sans Mono' # 'mono'
fontsize    = 10
fontcolor   = '#d5d2c8'
backcolor   = '#343d46'
activecolor = '#B43104'
cursorcolor = '#B43104'                 # activecolor
cursorwidth = 0                         # 0 = Blockcursor
cursor_ton =  1000 # in ms e.g 800      #  
cursor_toff = 0    # in ms e.g 200      # 0 = no blinking

in_string  =  "include"   # keyword for including external source files
in_string2 =  "#include"  # 2nd keyword for include
in_string3 =  "\ include" # 3rd keyword for include, can be overwritten

newstopperkey = "next line overrides default stopper" 

undo_depth = 256           # max. undo experimental
undo_index = 0             # index into the undo array
undo_text = []             # undo text
undo_curs = []             # undo cursor position
undo_scro = []             # undo scroll position
undo_text.append("\n")     # first entry is empty
undo_curs.append(0.0)      #
undo_scro.append(0.0)      #

# Default and global values -------------------------------------------

baud_string = "\ »baud:"  # keyword for baud rate switching
ok_string  =  "ok.\n"     # "ok." target answer
nok_string =  ".\n"       # not ok = any ".\n" target answer
mg_string  =  "»"         # magic string
logname    =  "log.txt"   # target output is logged here
helpfile   =  "FEE.org"   # help text

add_stopper_cr = True     # Stopper with linefeed
full_backslash = False    # True  = behind a \ is not send
logenable = False         # False = default is logging off
save_enable = True        # True  = default is backup copy active
save_path = "./sav"       #  
uart = True               # True  = default is serial Interface
uart_rtscts = False       # False = default is hardware handshake off
uart_dtr = False          # True = low; False = High, default: inactive = High 
uart_rts = True           # True = low; False = High, default: inactive = Low

tim_reload = 1000	  # waiting for char: tim_reload x tim_pause [ms]
tim_EOL    = 50           # waiting after EOL without anything else tim_EOL x tim_pause [ms] 
tim_pause  = 10.0         # time steps in timeout loop [ms]

telnet_timeout = 250.0    # 250s timeout to ping the host

filename  =  None         # start with no filename and some wise text :-)
filestate = "« mais penser ne tue pas »"

# storage for extra characters: «» 

# used libraries ------------------------------------------------------

import glob, time, os, os.path
import serial, serial.tools.list_ports

import telnetlib
from urllib.request import urlopen

if os.name == 'posix':
    import termios, sys, tty, select		# linux: non blocking input
else:
    import msvcrt, sys				# windows: non blocking input
    from tkinter.messagebox import showerror	# windows: needed

from tkinter import *
from tkinter import filedialog
from tkinter.filedialog import asksaveasfile
from tkinter.filedialog import askopenfile
from tkinter import messagebox
from sys import argv

# get the system initialized ------------------------------------------

def generate_initstring():	# read an initstring from command line
    initstr=argv		# copy the argument list and
    initstr.pop(0)              # delete the first element
    if len(initstr) < 2:        # if there are not enough arguments
        print('')
        print('» Hey, I need min. 2 parameters: device & baudrate or \'telnet\' & host')
        print('» serial : python FEE.py <interface> <baudrate> <options> ')
        print('» telnet : python FEE.py telnet <host>:<port> ')
        print('» for example: ')
        print('» python FEE.py /dev/ttyUSB0 115200 DTRN ')
        print('» python FEE.py telnet 192.168.4.1:23 ')
        print('»')
        print('» I can see following serial (USB) interfaces: ')
        ports = list(serial.tools.list_ports.comports())
        for p in ports:
            print ('»',p)
        print('')
        print('» For a detailed parameter list have a look into the manual,')
        print('» it is contained in the first several 100 code lines ;-) ')  
        exit(0)
    return(initstr)

def read_interface_parameters(initstr):
    global uart_port
    global uart_baud
    global uart_dtr
    global uart_rtscts
    global interface
    global telnet_host
    global uart                         # True/False = uart/telnet Interface
    uart_port   = initstr[0]		# the first 2 arguments determine  
    uart_baud   = initstr[1]		# port, baudrate, type, host
    interface   = initstr[0]		# 
    telnet_host = initstr[1]		#
    uart = not ("telnet" in interface)  #
    for i in range (0,len(initstr)):	# somewhere in the arguments
        try:				# might be a
            if 'DTRN' in string[i]:	# DTRN to invert DTR-Reset
                uart_dtr = True		#
            if 'RTSN' in string[i]:     # or a RTSN to invert RTS-Boot0
                uart_rts = False        #
            if 'RTSCTS' in string[i]:	# or a RTSCTS to activate
                uart_rtscts = True	# hardware handshake
        except:
            break
    return()  

def overload_defaults(initstr):
    global uart_dtr
    global uart_rts
    global full_backslash
    global save_enable
    global in_string3
    global fontname
    global fontsize
    global fontcolor
    global backcolor
    global activecolor
    global cursorcolor
    global cursorwidth
    global cursor_ton
    global cursor_toff
    global px_max
    global py_max
    global px_min
    global py_min
    global tx		# text area (unbenutzt?)
    global ty		#    
    for i in range (0,len(initstr)):
        try:
            if 'DTRN'   in initstr[i]: uart_dtr = not uart_dtr			# invert DTR logic
            if 'RTSN'   in initstr[i]: uart_rts = not uart_rts			# invert DTR logic
            if 'FASTER' in initstr[i]: full_backslash = True			# transfer minimum
            if 'NOSAVE' in initstr[i]: save_enable = False			# switch off backup copy
            if 'is=' in initstr[i]: in_string3  = (initstr[i].strip("is="))	# include string  
            if 'fn=' in initstr[i]: fontname    = (initstr[i].strip("fn="))	# font name
            if 'fs=' in initstr[i]: fontsize    = int(initstr[i].strip("fs="))	# font size
            if 'fc=' in initstr[i]: fontcolor   = (initstr[i].strip("fc="))	# font color
            if 'bc=' in initstr[i]: backcolor   = (initstr[i].strip("bc="))	# back color
            if 'ac=' in initstr[i]: activecolor = (initstr[i].strip("ac="))	# active color
            if 'cc=' in initstr[i]: cursorcolor = (initstr[i].strip("cc="))	# cursor color
            if 'cw=' in initstr[i]: cursorwidth = (initstr[i].strip("cw="))	# cursor width
            if 'co=' in initstr[i]: cursor_ton  = (initstr[i].strip("co="))	# cursor ton
            if 'cf=' in initstr[i]: cursor_toff = (initstr[i].strip("cf="))	# cursor toff
            if 'is=' in initstr[i]: in_string3  = (initstr[i].strip("is="))	# include string
            if 'tr=' in initstr[i]: tim_reload  = int(initstr[i].strip("tr="))  # time for reload
            if 'te=' in initstr[i]: tim_EOL     = int(initstr[i].strip("te="))  # time for EOL
            if 'tp=' in initstr[i]: tim_EOL     = int(initstr[i].strip("tp="))  # time for pause
        except:
            break
    px_max = fontsize * 70
    py_max = fontsize * 70
    px_min = fontsize * 40
    py_min = fontsize * 5
    tx = 100
    ty = 2000
    return()

def uart_init():			# init the serial interface
    global ser
    ser = serial.Serial()
    ser.port = uart_port		# eg. '/dev/ttyUSB0'
    ser.baudrate = int(uart_baud)	# eg. 115200
    ser.bytesize = serial.EIGHTBITS
    ser.parity = serial.PARITY_NONE
    ser.stopbits = serial.STOPBITS_ONE
    ser.timeout = 0
    ser.xonoff = False
    ser.rtscts = uart_rtscts
    ser.dsrdtr = False
    ser.writeTimeout = 0
    ser.open()
    ser.setDTR(uart_dtr)		# reset inactive
    ser.setRTS(uart_rts)		# boot0 inactive
    return()

def telnet_init():			# init the telnet connection
    global tn23
    global telnet_host
    global HOST
    global PORT
    global telnet_timer

    telnet_host = telnet_host.split(':')
    HOST = telnet_host[0]
    PORT = telnet_host[1]
    if PORT =="": PORT = '23'		# default is port 23
    tn23 = telnetlib.Telnet(HOST, PORT)
    tn23.open(HOST, PORT)
    telnet_timer = time.time() + telnet_timeout
    return()

def init_interface():			# init interface, which is used 
    global uart
    if uart: uart_init()
    else:  telnet_init()
    return()

# communication functions ---------------------------------------------

# timeout function for telnet connection avoids disconnect from host,
# like some WiFi adapters do 

def reload_telnet_timer(): 
    global telnet_timer
    telnet_timer = time.time() + telnet_timeout
    return()

def test_timer():
    if uart == False:
        global telnet_timer
        if time.time() > telnet_timer:
            reload_telnet_timer()
            tn23.write(str.encode('\0'))
            print('[ping]')
    return()

def uart_empty():			# get all buffers clean
    global ser
    ser.reset_input_buffer()
    ser.reset_output_buffer()
    return()

def byte_read_serial():			# reading from serial port
    global ser
    try:
        c = (ser.read(ser.in_waiting)).decode("utf-8","ignore")
    except:
        c=''
        save_edit()
        print('\n» UART error! «')
        ser.reset_input_buffer()
    return(c)

def byte_read_telnet():			# reading from telnet
    c = (tn23.read_eager()).decode("utf-8","ignore")
    # reload_telnet_timer()
    return(c)

def byte_read():			# reading depending on active interface
    if uart == True:
        c = byte_read_serial()
    else:
        c = byte_read_telnet()
    return(c)

def read_target():
    c = byte_read()
    if len(c) != 0:
        if logenable: write_log(c)
        c = c.replace('\r', '\n')            # line end converted from \r to \n
        sys.stdout.write(c)                  # output to terminal
        sys.stdout.flush()                   # without waiting
    return(c)

def write_target(c):
    if uart == True:
        if os.name == 'posix':
            ser.write(bytes(c,'utf-8'))      # linux uses utf-8
        else:
            ser.write(bytes(c))              # windows not
    else:
        tn23.write(str.encode(c))
    return(c)

# editor functions ----------------------------------------------------

def load_config():
    line = read_line(1)						# 1st line = config line
    if ( '(' in line ) and ( ')' in line ) or '\\' in line:	# valid comment parts?
        initstr = line.split(' ')
    overload_defaults(initstr)
    return()

def load_stopper():			                       # if magic word in 1st line,
    global STOPPER              	                       # use 2nd line stopper template
    line = read_line(1)                                        # 1st line = config line
    if  ( '(' in line ) and ( ')' in line ) or '\\ ' in line:  # vaild comment parts?

        if newstopperkey in line:                              # if it contains stopper magic
            line = read_line(2)                                # read 2nd line and check
            if ( '(' in line ) and ( ')' in line ) or '\\ ' in line:
                STOPPER = read_line(2)                         # load stopper template
    load_config()
    return()

def set_stopper():
    editortext.insert("insert", ('\n'+STOPPER))
    return()

def scan_stopper():
    pos_edit = (editortext.index(INSERT))
    editortext.mark_set("mark_edit", index_add(pos_edit, 0, 1))
    editortext.mark_gravity("mark_edit", "left")
    row, col = pos_edit.split('.')	# watch out: this patches the read position to
    row = str( int(row) + 1 ) 		# one line beyond to mark block below cursor
    col = str(0)                        # for stopper scan the column 0
    pos_edit = row + '.' + col		# on the stopper line
#    pos_scan1 = editortext.search(STOPPER[0:(stonu-1)] ,pos_edit, stopindex="1.0", backwards=1)
    pos_scan1 = editortext.search(STOPPER, pos_edit, stopindex="1.0", backwards=1) 
    if pos_scan1 == "": pos_scan1 = "0.0"
    editortext.mark_set("mark_scan1", index_add(pos_scan1, 0, 1))
    editortext.mark_gravity("mark_scan1", "left")
#    pos_scan2 = editortext.search(STOPPER[0:(stonu-1)] ,pos_edit, stopindex="end")
    pos_scan2 = editortext.search(STOPPER, pos_edit, stopindex="end")
    if pos_scan2 == "": pos_scan2 = editortext.index("end")
    editortext.mark_set("mark_scan2", index_add(pos_scan2, 0, 1))
    editortext.mark_gravity("mark_scan2", "left")
    return()

def index_add(index, drow, dcol):
    row, col = index.split('.')
    row = str( int(row) + drow)
    col = str( int(col) + dcol)
    index = row + '.' + col
    return(index)

def insert_line(line, text):
    index1 = str(line) + '.0'
    editortext.insert(index1, text)
    #editortext.insert(index1, text + '\n')

def read_line(line):
    index1 = str(line) + '.0'
    index2 = str(line+1) + '.0'
    #index2 = str(line) + '.end'
    text = editortext.get(index1, index2)
    return(text)

def delete_line(line):
    index1 = str(line) + '.0'
    index2 = str(line+1) + '.0'
    #index2 = str(line) + '.end'
    text = editortext.delete(index1, index2)

def get_cursor_line():
    position = editortext.index(INSERT)
    row, col = position.split('.')
    line = int(row)
    return(line)

def stopper_line(stopper):
    position = editortext.index(stopper)
    row, col = position.split('.')
    line = int(row)
    return(line)

def upper_stopper_line(): return( stopper_line("mark_scan1"))
def lower_stopper_line(): return( stopper_line("mark_scan2"))

def erase_line():
    global line_buffer
    line = get_cursor_line()
    editor.clipboard_clear()
    editor.clipboard_append(read_line(line))
    delete_line(line)
    return()

def erase_line_event(event):
    restore_space_if()
    erase_line()

def save_line():
    global line_buffer
    line = get_cursor_line()
    editor.clipboard_clear()
    editor.clipboard_append(read_line(line))
    return()

def save_line_event(event):
    restore_space_if()
    save_line()

def cut_textbox():
    if editor.selection_get():
        textbox = editor.selection_get()
        editor.clipboard_clear()
        editor.clipboard_append(textbox)
        text = editortext.delete('sel.first','sel.last')
    return()

def copy_textbox():
    if editor.selection_get():
        textbox = editor.selection_get()
        editor.clipboard_clear()
        editor.clipboard_append(textbox)
    return()

def paste_textbox():
    text = editor.clipboard_get()
    index = editortext.index(INSERT)
    editortext.insert(index, text)
    return()

def include_filename_line(zeile,string):
    if zeile[0:len(string)] == string:
        zeile = zeile[len(string):-1]
        zeile = zeile.split('\\ ')
        zeile = zeile[0]
        zeile = zeile.strip()
    else:
        zeile=""
    return(zeile)

def include_filename(zeile):
    name = include_filename_line(zeile, in_string)
    if name =="":
        name = include_filename_line(zeile, in_string2)
        if name =="":
            name = include_filename_line(zeile, in_string3)
    if name !="": print('\n» include: ', name, '\n')
    return(name)

def watch_baud(zeile):
    if baud_string in zeile:
        zeile = zeile.split(baud_string)
        baud = zeile[1]
        baud = baud.strip()
    else:
        baud =''
    return(baud)

def switch_baud(baud):
    global ser
    if baud != "":
        if uart:
            ser.baudrate = int(baud)
        else:
            x = urlopen('http://' + HOST + '/console/baud?rate=' + baud)
        print('\n»new baudrate: ', baud, '\n')
    return() 

def read_ok():						# read from target, return ok/nok
    ok = False						# default nok
    i = tim_reload					# in a loop starting with 1000
    while i > 0:					# max. tim_reload times
        c = byte_read()					# read target
        if len(c) !=0:					# something read
            i = tim_reload                              # reset timeout
            sys.stdout.write( c.replace('\r', '\n') )	# write to terminal with modified EOL
            sys.stdout.flush()				# without delay
            if ok_string in c:  		        # ok.\n seen?
                if logenable: write_log(c)	        # write to log
                ok = True				# remeber ok
                break					# leave loop with ok
            else:					# no ok. seen
                if nok_string in c:			# any .\n seen?
                    break				# leave loop with nok

            if '\n' in c: i = tim_EOL                   # if EOL short timeout

        else:					        # nothing read
            time.sleep(tim_pause / 1000.0)              # pausing
            i = i - 1                                   #
    return(ok)

def reduziere_zeile(zeile):
    zz = zeile
    if (zz[0:2] == '\\ ') or (zz[0:2] == '\\\n'):	# for lines with \ at first position
        zz = ''						# generate an empty skipping line
    else:						# if not a \ at 1st position
        if full_backslash:				# maybe further reduction
            zz = (zz[0:-1])                             # first delete EOL
            if zz[-1] == '\\' : zz = zz[0:-1]           # remove trailing \ if trailing
            if ' \\ ' in zz:				# looking for the next \ , if
                zz = zeile.split(' \\ ')		# split at \ and 
                zz = zz[0]				# take only text in front of \
            zz = zz.strip()                             # strip leading and trailing spaces
            zz = zz + '\n'			        # and append a new EOL
    return(zz)

def schreibe_zeile(zeile):				# write a line and check for "ok."
    baud = watch_baud(zeile)                            # before changing line look for baud
    zeile = reduziere_zeile(zeile)                      # reduce/skip line
    if len(zeile) > 0:                                  # 
        if uart == True:                                # send for uart
            for i in range (0 , len(zeile)):            # byte by byte
                ser.write(bytes(zeile[i],'utf-8'))	# how does it work at windows?
            switch_baud(baud)                           # switch serial baudrate on request
        else:                                           # send for telenet
            tn23.write(str.encode(zeile))               # the whole line
            reload_telnet_timer()                       # reload the telnet timeout timer
            switch_baud(baud)                           # switch baudrate at esp-link
        ok = read_ok()                                  # and read answer
    else:                                               # empty line no send
        ok = True                                       # but everything is ok
    return(ok)

def compiliere_zeile(zeile):			# send a line for compilation
    ok = True					# default = ok
    if zeile != '\n':				# only if line is not empty
        text = include_filename(zeile)          # check for include and return filename
        if text == '':                          # with no filename it's a simple line
            if not '\n' in zeile:               # if the line does not contain an end,
                zeile = zeile + '\n'            # e.g. last line in include file, add an end
            ok = schreibe_zeile(zeile)          # to write to the target
        else:                                   # with filename
            try:                                # it could be an include file
                with open(text) as f:           # open file
                    errline = 0                 # error line number                      
                    for line in f:              # send it line by line
                        errline = errline + 1   # increment error line number
                        ok = compiliere_zeile(line)
                        if ok == False:         # if an error occurs
                            print("» ERROR IN INCLUDE FILE ",text," LINE ",errline," «")
                            break               # halt the transfer    
            except:                             # something went wrong with the file
                print("» INCLUDE FAILED «")     # show it immediatly on terminal
                ok = False                      # and set flag to not ok
    return(ok)                                  # return with success flag

def transfer_block():                           # transfer a block
    ok = True                                   # assume all is ok
    scan_stopper()                              # locate the stoppers
    line1 = upper_stopper_line() + 1            # one line below stopper
    line2 = lower_stopper_line()                # should be a line above stopper
    for line in range (line1, line2):           # work line by line
        text = read_line(line)                  # take a line
        ok = compiliere_zeile(text)             # send for compiling, on error ok=False
        if ok == False:                         # check for an error
            print("» ERROR IN LINE ",line," «") #
            break                               # leave the loop
    return(ok)                                  #

def process_lines():                            # process lines / one block
    save_edit()                                 # right before changing, make a backup
    global filestate                            # status line message
    filestate='compile ok'                      # assume first it'll be ok
    ok = transfer_block()                       # transfer the actual block
    if ok == False:                             # something went wrong
        print("» COMPILING STOPPED «\n")        # error message into terminal
        filestate ='compile failed'             # and error message into status line        
    else:                                       # it went right
        scan_stopper()                          # to be sure
        delete_line(upper_stopper_line())       # move upper stopper down
        insert_line(lower_stopper_line(),(STOPPER+'\n'))
    return(ok)                                  #    

def try_lines():                                # transfer a block as trial
    global filestate                            # status line message
    filestate='trial ok'			# assume first it'll be ok
    ok = transfer_block()                       # transfer the actual block
    if ok == False:                             # something went wrong
        print("» TRIAL STOPPED «\n")            # error message into terminal
        filestate ='trial failed'               # and error message into status line
    return(ok)                                  #

def try_a_line():				# try a single line
    global filestate				# filestate for	
    filestate='line trial ok'			# status line at first ok	
    ok = True                                   # default state is ok
    line = get_cursor_line()			# get the current position 
    zeile = read_line(line)			# get the line
    ok = compiliere_zeile(zeile)		# send the line for compiling
    if ok == False:				# check for an error
        filestate ='line trial failed'		# error message into status line
    return()   

# file operations -----------------------------------------------------

def newfile():					# erase text area
    global STOPPER				# reset STOPPER
    STOPPER = STOPPER_DEF                       # default
    global filename				#
    global filestate				#
    save_edit()					# do a backup befor
    editortext.delete(0.0, END)                 # deleting text area
    filename = "noname.txt"			# some default file name
    filestate = "new file"			# message to statusbar
    return()

def writefile(filename, text):			# write file
    global filestate				# 
    save_edit()					# do a backup
    try:					# 
        if filename == None:			# witho no filename ask for a name
            file = asksaveasfile(defaultextension='.txt')
            filename = file.name		# take the filename
        f = open(filename, 'w')			# open file to write
        f.write(text)				# write
        f.close()				# and close
    except Exception as e:			# if there was an error
        #messagebox.showerror("Exception",e)	#
        filestate = "file write failed"		# pass an error message to the status
    return(filename)

def savefile():					# action for saving a file
    global filename				#
    global filestate				#
    filestate = "file saved"			# default is no error
    filename = writefile(filename, editortext.get(0.0, END))
    return()					# use writefile for writing

def savefile_event(event):			# the key event save file
    savefile()

def savefileas():				# action saving as
    global filename				#	
    global filestate				#
    filestate = "file saved as"			# default is no error
    filename = writefile(None, editortext.get(0.0, END))
    return()					# writefile with noname does it

def savefileas_event(event):			# the key event for save as
    savefileas()

def magicblock():				# save a text block and replace it 
    global filestate
    filestate = "block replaced"		# default status, if no errors
    filename = None				# filename has to be generated later
    scan_stopper()                              # watch out the STOPPER positions
    pos1 = editortext.index("mark_scan1")       # upper stopper
    row1, col1 = pos1.split('.')                # get line number
    pos2 = editortext.index("mark_scan2")       # lower stopper
    row2, col2 = pos2.split('.')                # get line number
    zeile = read_line(int(row1)+1)              # read 1st line of block
    if (zeile[0:1] == mg_string):               # is there a magic string?
        filename = zeile[1:-1]                  # if so read rest of line as filename
        if filename =='': filename = None       # if no filename is given make it none
        savpos1 = str(int(row1)+2) + '.0'       # begin of first line to save
        savpos2 = str(int(row2)-1) + '.end'     # end of last line to save
    else:                                       # if there is no magix string
        savpos1 = str(int(row1)+1) + '.0'       # begin of first line to save
        savpos2 = str(int(row2)-1) + '.end'     # end of last line to save
    filename = writefile(filename, editortext.get(savpos1,savpos2)) # save the block
    if (filename == None) or (filename==""):	# if generating a filename failed
        filestate = "block failed"		# show it in the status line
    else:                                       # if saving was ok
        delpos1 = str(int(row1)) + '.0'         # begin of first line to erase
        delpos2 = str(int(row2)) + '.end'       # end of last line to erase
        editortext.delete(delpos1, delpos2)     # and erase
        text = in_string + ' ' + filename       # inculde string + filename
        editortext.insert(delpos1, text)        # put it into the text
    return()

def magicblock_event(event):                   # key event for replacing the block
    magicblock() 

def magicmark_event(event):                    # key event to put the magic string
    editortext.insert(editortext.index(INSERT), mg_string)

def readfile():
    global filename
    global filestate
    save_edit()
    try:
        file = askopenfile(parent=editor,title='select file')
        filename = file.name
        t = file.read()
        file.close()
        filestate = "file read"
        editortext.delete(0.0, END)
        editortext.insert(0.0, t)
        editortext.delete('end - 1 chars')
        load_stopper()
        scan_stopper()
        save_edit()
    except:
        filestate = "file read failed"
    return()

def readfile_event(event):
    readfile()

def reloadfile():
    global filename
    global filestate
    if filename == None:
       filestate = "no file to reload" 
    else:
        # try:
            file = open(filename, 'r')          # open
            t = file.read()                     # read
            file.close()                        # close
            filestate = "file reloaded"		# prepare status line
            editortext.delete(0.0, END)		# empty editor area
            editortext.insert(0.0, t)		# insert text from file
            editortext.delete('end - 1 chars')  # delete last character
            load_stopper()                      # read stopper template from 1st line
            scan_stopper()			# look for stopper positions
        # except:
        #    filestate = "file reload failed"
    return()

def reloadfile_event(event):
    reloadfile()

def save_copy(text):			                   # backup copy if not empty text area
    if ( ( not text == "" ) and ( not text =="\n") and (save_enable == True) ):
        savefile = (time.strftime('%Y-%m-%d_%H-%M-%S') + '.sav')  # backup filename
        if os.name == 'posix':                                    # linux only supports backup path
            if not os.path.isdir(save_path):                      # is there no backup directory
                try:                                              # then try to 
                    os.mkdir(save_path)                           # create the directory
                    savefile = save_path + '/' + savefile         # adjust the filename
                except:                                    
                    print('\n» Error: Could not create backup directory! «')
            else:                                                 # if there is already the directory
                savefile = save_path + '/' + savefile             # adjust only the filename
        # store a backup with or without separate path for backup files
        try:
            f = open( savefile, 'w')
            f.write(text)
            f.close()
        except Exception as e:
            #messagebox.showerror("Exception",e)
            print('\n» Error: Could not write backup copy! «')
    return()

def save_edit():			 # make a backup copy of text area
    restore_space_if()			 # convert space if there were dots
    save_copy(editortext.get(0.0, END))	 # save
    return()

def quit_program():
    global nochmal
    #editor.quit()
    nochmal = False
    return()

def quit_event(event):
    quit_program()

def DTR_reset():
    if uart:							# for the uart:
        print('\n» DTR-RESET  ¡Hasta la vista, baby! «')	# reset message
        ser.setDTR(not uart_dtr)				# activate reset
        time.sleep(0.2)						# wait some time
        ser.setDTR(uart_dtr)					# deactivate reset
        time.sleep(0.1)						# wait some time before
        ser.baudrate = int(uart_baud)				# set default baudrate
        print('» Baudrate reset to', uart_baud, ' «\n')		# give a hint
    else:							# for telenet:
        print('\n» REMOTE-RESET  ¡Hasta la vista, baby! «')	# reset message
        x = urlopen('http://' + HOST + '/console/reset')	# reset order to esp-link
    return()

def DTR_reset_event(event):
    DTR_reset()

def clearoutput():
    restore_space_if()
    if os.name == 'posix':
        _ = os.system('clear')
    else:
        _ = os.system('cls')
    return() 

def clearoutput_event(event):
    clearoutput()

# editor operations ---------------------------------------------------

# def tab_event(event):                          # key event TAB
#     editortext.insert(editortext.index(INSERT), '  ')

def trylines_event(event):
    restore_space_if()
    try_lines()

def processlines_event(event):
    restore_space_if()
    process_lines()

def patch_delete_CR():				# patch: delete CR after ALT/CTRL/SHIFT-Return
    pos_ed = (editortext.index(INSERT))		# position where CR is inserted
    row, col = pos_ed.split('.')		# split into row and column
    row = str( int(row) - 1 )			# one line backwards
    col = '10000'                               # workaraound: row = end
    pos_de = row + '.' + col			# build (position -1)
    editortext.delete(pos_de, pos_ed)		# delete
    return()

def trylines_event_CR(event):
    patch_delete_CR()
    restore_space_if()
    try_lines()

def try_a_line_event_CR(event):
    patch_delete_CR()
    restore_space_if()
    try_a_line()

def processlines_event_CR(event):
    patch_delete_CR()
    restore_space_if()
    process_lines()

def setstopper_event(event):
    restore_space_if()
    set_stopper()

def open_log():
    global logname
    global logobj
    logobj = open(logname, "a")
    logobj.write("\n[" + (time.strftime('%Y.%m.%d %H:%M:%S')) + "]»\n") 
    
def close_log():
    global logobj
    logobj.write("\n«[" + (time.strftime('%Y.%m.%d %H:%M:%S')) + "]\n")
    logobj.close()
        
def write_log(text):
    global logobj
    logobj.write(text)

def toggle_log():
    global logenable
    if logenable :
        logenable = False
        close_log()
        #print("\n[LOG OFF]\n")
    else:
        logenable = True
        open_log()
        #print("\n[LOG ON]\n")

def toggle_log_event(event):
    toggle_log()

space_marker = False # default is no spaces not marked as dots

def replace_space():
    ttt = editortext.get('1.0','end')
    ttt = str.replace(ttt, ' ', '·')
    editortext.delete(0.0, END)
    editortext.insert(0.0, ttt)
    editortext.delete('end - 1 chars')
    return()

def restore_space():
    ttt = editortext.get('1.0','end')
    ttt = str.replace(ttt, '·', ' ')
    editortext.delete(0.0, END)
    editortext.insert(0.0, ttt)
    editortext.delete('end - 1 chars')
    return()

def restore_space_if():
    global space_marker
    if space_marker:
        space_marker = False
        restore_space()
    return()

def toggle_space():
    pos_edit = (editortext.index(INSERT))
    poss = str(editortext.yview())
    poss = poss.split(',')
    poss = poss[0]
    poss = poss[1:]
    global space_marker
    if space_marker:
        space_marker = False
        restore_space()
    else:
        space_marker = True
        replace_space()
    #print(space_marker)
    editortext.mark_set(INSERT,pos_edit)
    editortext.focus_set()
    editortext.yview_moveto(poss)
    return()

def toggle_space_event(event):
    toggle_space()

def positionp(position, delta_col ):
    row, col = position.split('.')
    col = str(int(col) + delta_col)
    position = row + '.' + col
    return(position)

def test_stopper():
    restore_space_if()
    editortext.tag_delete('tag')
    scan_stopper()
    position = editortext.index("mark_scan1")
    editortext.tag_add('tag', positionp(position, +1), positionp(position, stonu))
    editortext.tag_config('tag', background=backcolor, foreground=activecolor)
    position = editortext.index("mark_scan2")
    editortext.tag_add('tag', positionp(position, +1), positionp(position, stonu))
    editortext.tag_config('tag', background=backcolor, foreground=activecolor)
    return()

def test_stopper_event(event):
    test_stopper()

def get_word_at_cursor():
    position = editortext.index(INSERT)
    row, col = position.split('.')
    row = int(row)
    col = int(col)
    t = read_line(row)
    i = 0 
    for i in range (col, len(t)):
        if t[i] == ' ' : break
    t = t[:i]
    for i in range (0, len(t)):
        if t[(len(t)-i)-1] == ' ' : break
    t = t[(len(t)-i-1):]
    t = t.strip()    
    return(t)

def F1():
    w = get_word_at_cursor()
    try:
        help.deiconify()
        help.update()
        help_read()
        if w != '' : help_search(w)
    except:
        pass

def F1_event(event):
    F1()

def target_flash():
    try:
        target_load()
        uart_init()
    except:
        pass
    return()

def Status_update():
    c = " " + editortext.index(INSERT) + " |"
    if logenable: c = c + "LOG|"
    c = c + " " + filestate
    if filename: c = c + " | " + filename
    c = c + " | " + str(undo_index)
    editor.status.set(c)
    return()

def scrollposition():
    poss = str(editortext.yview())                  # read scroll position
    poss = poss.split(',')                          # extract position
    poss = poss[0]                                  # 1st part
    poss = poss[1:]                                 # without 1st char
    return(poss)

def undo_add_event(key):
    global undo_index
    global undo_depth
    global undo_text
    global undo_curs
    global undo_scro
    text = editortext.get('1.0','end')			# text in buffer
    if undo_text[undo_index] != text :                  # only if text is new
        if len(undo_text)-1 > undo_index :		# is index moved down?
            delta = undo_index - len(undo_text) + 1     # then delete
            del undo_text[delta:]                       # old future
            del undo_curs[delta:]                       #
            del undo_scro[delta:]                       #
        undo_text.append(text)				# add text
        undo_curs.append(editortext.index(INSERT))	# add cursor position
        undo_scro.append(scrollposition())        	# add scrollposition
        if len(undo_text) > undo_depth :                # max. depth ?
            del undo_text[0]				# delete oldest entry
            del undo_curs[0]				#
            del undo_scro[0]				#
        undo_index = len(undo_text)-1                   # sync index
    return()

def undo():
    global undo_index
    global undo_depth
    global undo_text
    global undo_curs
    global undo_scro
    if undo_index > 0 :						# lower limit for undo
         undo_index = undo_index - 1				# decrement index
         editortext.delete(0.0, END)				# empty text area
         editortext.insert(0.0, undo_text[undo_index])		# restore old text
         editortext.delete('end - 1 chars')			# delete last CR
         editortext.mark_set(INSERT,undo_curs[undo_index])	# restore cursor position
         editortext.yview_moveto(undo_scro[undo_index])		# restore scroll position
         editortext.focus_set()					# focus on editor
    return()

def redo():
    global undo_index
    global undo_depth
    global undo_text
    global undo_curs
    global undo_scro
    if undo_index < (len(undo_text) - 1) :
        print
        undo_index = undo_index + 1
        text = undo_text[undo_index]
        editortext.delete(0.0, END)
        editortext.insert(0.0, text)
        editortext.delete('end - 1 chars')
        editortext.mark_set(INSERT,undo_curs[undo_index])
        editortext.yview_moveto(undo_scro[undo_index])
        editortext.focus_set()
    return()

def undo_event(event):
    undo()

def redo_event(event):
    redo()

# get the system started ----------------------------------------------
 
initstr = generate_initstring()
string = read_interface_parameters(initstr)
overload_defaults(initstr)
init_interface()

## build an editor window with tkinter

editor = Tk()

editor.title(KopfZeile)
editor.minsize(width=px_min, height=py_min)
editor.maxsize(width=px_max, height=py_max)
editor.geometry("820x600+550+150")

editor.status = StringVar()
editorstatus = Label(editor, textvariable = editor.status, anchor=W, bd=1)
editorstatus.pack(side=BOTTOM, fill=X)
editor.status.set("")

editortext = Text(editor, takefocus=1, font=(fontname,fontsize), bg=backcolor, fg=fontcolor, insertbackground=cursorcolor, state="normal", width=tx, height=ty, blockcursor=True, insertwidth=cursorwidth, insertontime=cursor_ton, insertofftime=cursor_toff)
editortext.pack(side=LEFT,fill=BOTH, expand=1)

editorscroll = Scrollbar(editortext, orient=VERTICAL)
editorscroll.config(command = editortext.yview)
editorscroll.pack(side=RIGHT, fill=Y)

editortext.config(yscrollcommand = editorscroll.set)

# editor menus #

#Menuzeile = Menu(editor, font=(fontname,fontsize), bg=backcolor, fg=fontcolor )
Menuzeile = Menu(editor, bd=1)

Dateimenu = Menu(Menuzeile)
Dateimenu.add_command(label="file new",       command=newfile )
Dateimenu.add_command(label="file open",      command=readfile,     accelerator = 'ctrl+o')
Dateimenu.add_command(label="file reload",    command=reloadfile,   accelerator = 'ctrl+r')
Dateimenu.add_command(label="file save",      command=savefile,     accelerator = 'ctrl+s')
Dateimenu.add_command(label="file save as",   command=savefileas,   accelerator = 'ctrl+shift-s')
Dateimenu.add_command(label="log on/off",     command=toggle_log,   accelerator = 'ctrl+shift+w')
Dateimenu.add_separator()
Dateimenu.add_command(label="quit",           command=quit_program, accelerator = 'ctrl+shift+q')
Dateimenu.add_separator()
Menuzeile.add_cascade(label="file", menu=Dateimenu)

Arbeitsmenu = Menu(Menuzeile)
Arbeitsmenu.add_command(label="undo",     command=undo,          accelerator = 'ctrl+z')
Arbeitsmenu.add_command(label="redo",     command=redo,          accelerator = 'ctrl+shift+z')
Arbeitsmenu.add_separator()
Arbeitsmenu.add_command(label="cut",      command=cut_textbox,   accelerator = 'ctrl+x')
Arbeitsmenu.add_command(label="copy",     command=copy_textbox,  accelerator = 'ctrl+c')
Arbeitsmenu.add_command(label="paste",    command=paste_textbox, accelerator = 'ctrl+v')
Arbeitsmenu.add_separator()
Arbeitsmenu.add_command(label="line cut",   command=erase_line,      accelerator = 'ctrl+l')
Arbeitsmenu.add_command(label="line copy",  command=save_line,       accelerator = 'alt+l')
Arbeitsmenu.add_command(label="line paste", command=paste_textbox,   accelerator = 'ctrl+v')
Arbeitsmenu.add_separator()
Arbeitsmenu.add_command(label="line try",   command=try_a_line,    accelerator = 'shift+return')
Arbeitsmenu.add_command(label="block try",  command=try_lines,     accelerator = 'ctrl+return')
Arbeitsmenu.add_command(label="block send", command=process_lines, accelerator = 'alt+return')
Arbeitsmenu.add_separator()
Arbeitsmenu.add_command(label="stopper insert",   command=set_stopper,   accelerator = 'ctrl+j')
Arbeitsmenu.add_command(label="block convert",    command=magicblock,    accelerator = 'alt+j')
Arbeitsmenu.add_separator()
Arbeitsmenu.add_command(label="spaces hide/show", command=toggle_space,  accelerator = 'ctrl+shift+m')
Arbeitsmenu.add_separator()
Menuzeile.add_cascade(label="edit", menu=Arbeitsmenu)

Werkzeugmenu = Menu(Menuzeile)
Werkzeugmenu.add_command(label="terminal clear",     command=clearoutput,  accelerator = 'ctrl+q')
Werkzeugmenu.add_separator()
Werkzeugmenu.add_command(label="target reset",       command=DTR_reset,    accelerator = 'alt+q')
if uart == True:
    Werkzeugmenu.add_separator()
    Werkzeugmenu.add_command(label="firmware flash", command=target_flash)
Werkzeugmenu.add_separator()
Menuzeile.add_cascade(label="tools", menu=Werkzeugmenu)

editor.config(menu=Menuzeile) 

# editor hot-keys

editor.bind("<Control-o>", readfile_event)
editor.bind("<Control-r>", reloadfile_event)
editor.bind("<Control-s>", savefile_event)
editor.bind("<Control-S>", savefileas_event)
editor.bind("<Control-W>", toggle_log_event)
editor.bind("<Control-Q>", quit_event)

editor.bind("<Control-l>", erase_line_event)
editor.bind("<Alt-l>",     save_line_event)

editor.bind("<Control-Return>", trylines_event_CR)
editor.bind("<Shift-Return>",   try_a_line_event_CR)
editor.bind("<Alt-Return>",     processlines_event_CR)

editor.bind("<Control-j>", setstopper_event)
editor.bind("<Control-T>", test_stopper_event)

editor.bind("<Alt-j>",     magicblock_event)
editor.bind("<Alt-m>",     magicmark_event)

editor.bind("<Control-M>", toggle_space_event)

editor.bind("<Control-q>", clearoutput_event)
editor.bind("<Alt-q>",     DTR_reset_event)

editor.bind("<Control-u>", undo_event)
editor.bind("<Control-z>", undo_event)
editor.bind("<Control-Z>", redo_event)
editor.bind("<KeyPress>", undo_add_event)
editor.bind("<KeyRelease>", undo_add_event)

editor.bind("<F1>", F1_event)

# editor functions
# »-------------------------------------------------------------------«
# flash functions

target_addr = 0x08000000	# code address
target_file = 'target.bin'	# default source file name
target_copy = 'target.sav'	# default backup file  
#data = []			# global data block
blen = 256			# default block length
bcnt = 512                      # default max. block count
wc = 0				# write/erase cmd


# serial interface init for flash-loader

read_char   = '«'  
write_char  = '»' 
verify_char = '#'
# « »  

def pling(zeichen):
    sys.stdout.write(zeichen)	# output to terminal
    sys.stdout.flush()		# without waiting

def uart_init_loader():
    global ser
    ser          = serial.Serial()
    ser.port     = uart_port
    ser.baudrate = int(uart_baud)
    ser.bytesize = serial.EIGHTBITS
    ser.parity   = serial.PARITY_EVEN
    ser.stobits  = serial.STOPBITS_ONE
    ser.timeout  = 2  # 2
    ser.xonoff   = False
    ser.rtscts   = False
    ser.dsrdtr   = False
    ser.writeTimeout = 0
    ser.open()
    ser.setDTR(uart_dtr)
    ser.setRTS(uart_rts)
    return()


def target_ack():
    try:
        ack = ord(ser.read())
#        ack = ord(ser.read(ser.in_waiting))
    except:
        ack =''
        print(' no response from target! ')
    else:
        if ack == 0x79 : return(True)
        else :
            print('no ack:', hex(ack))
            return(False)
    return(False)

def target_ack_():
    ack = ord(ser.read())
    return(ack == 0x79)

def target_reset():			# reset with Bootx pulled for UART Bootmode

    ser.setRTS(not uart_rts)            # pull BOOT0 active
    time.sleep(0.1)                     # 100ms delay

    ser.setDTR(not uart_dtr)		# pull RESET active
    time.sleep(0.1)			# 100ms delay

    ser.setDTR(uart_dtr)     		# release RESET
    time.sleep(0.1)    			# 100ms delay

    ser.setRTS(uart_rts)       		# release BOOT0
    time.sleep(0.5)                     # 500ms delay befor sending something

    ser.write(b'\x7F')			# send start byte (STM:AN3155 p.5/42)
    return(target_ack())		# wait for ack


def target_get():			# Command x00 to get Version & allowed commands
    global wc				# remember write/erase cmd type
    ser.write(b'\x00')			# send cmd x00
    ser.write(b'\xFF')			# plus crc
    if target_ack():			# wait for ack, if ack go on
        n = ord(ser.read())		# number of bytes
        v = ord(ser.read())		# version
        w = 0				# memory for w cmd 
        for i in range (0, n):		# read all cmds
            c = ord(ser.read())		# cmd
            if c == 0x44 : w = c	# remember x044
            if c == 0x43 : w = c	# remember x043 
        if target_ack():		# if ack, the read is valid
            print( hex(v), end = ' ' )	# print version
            wc = w			# finalise erase cmd
            return(True)		# done 
    return(False)			# all other not done   


def target_addrblk(address):		# build address block plus crc
    blk = []				# empty block
    b3 = (address >>  0) & 0xFF		# LSB
    b2 = (address >>  8) & 0xFF		#
    b1 = (address >> 16) & 0xFF		#
    b0 = (address >> 24) & 0xFF		# MSB
    crc = b0 ^ b1 ^ b2 ^ b3		# crc
    blk.append(b0)			# combine to block
    blk.append(b1)			#
    blk.append(b2)			#
    blk.append(b3)			#
    blk.append(crc)			#
    return(blk)				# return the block


def target_read_block(addr, blen):		# read and return a block from target
    block = []					# empty block
    ser.write(b'\x11')				# write cmd 0x11
    ser.write(b'\xEE')				# and crc
    if target_ack():				# wait for ack
        ser.write(target_addrblk(addr)) 	# write address block with crc
        if target_ack():			# if ack go on
            l = blen - 1 & 0xFF			# length byte = length-1 
            c = blen - 1 ^ 0xFF & 0xFF		# crc of lengh byte
            block.append(l)			# build length block
            block.append(c)			# with crc
            ser.write(block)			# write to target	
            if target_ack():			# if ack go on
                block = (ser.read(blen))	# read the data
                return(block)			# return the block
    return()					# all others return nothing


def target_read(addr, bcnt, blen):		# read bcnt*blen bytes from addr an copies it into a file
    ffbl = 4					# empty flash blocks for stop reading
    data = []					# empty block
    print('\n  read  :', end='')		# output
    for n in range (0, bcnt):			# read max. bcnt blocks
        block = target_read_block(addr,blen)	# read a block
        if len(block) != blen :			# if no block read
            print(' read failed ')		# it's a failure	
            return(False)			# and return with no success
        addr = addr + blen			# go on
        c = 0xFF				# check for all block bytes = 0xFF
        for i in range (0, len(block)-1):	# by running through the block
            c = c & block[i]			# and ANDing all bytes
        if c == 0xFF:				# if all = 0xFF
            ffbl = ffbl - 1			# decrement FF counter on empty block
            if ffbl == 0:                       # 
                break			        # stop reading, because of empty flash
        else:				        # if not empty
            ffbl = 4				#
        data.append(block)		        # append block to data
        pling(read_char)			# sign of life
    try:					# try to open & save
        f = open(target_copy, 'wb')		# open
        for b in data:				# write block by block
            f.write(b)				# write
        f.close()				# and close
    except:					# if not written
        print(' save failed ')			# show up a message
        return(False)				# and return with no success
    return(True)				# otherwise return with success


def target_global_erase():
    if wc == 0x43 :				# if cmd 0x43 supported
        ser.write(b'\x43')			# send cmd 0x43
        ser.write(b'\xBC')			# and it's crc
        if target_ack():			# if ack go on
            ser.write(b'\xFF')			# global erase
            ser.write(b'\x00')			# with crc
            if target_ack():			# if ack go on
                return(True)			# done
        return(False)				# all other is fail
    if wc == 0x44 :				# if cmd 0x44 supported
        ser.write(b'\x44')			# send cmd 44
        ser.write(b'\xBB')			# and it's crc
        if target_ack():			# if ack go on
            ser.write(b'\xFF')			# global erase
            ser.write(b'\xFF')			#
            ser.write(b'\x00')			# with crc
            if target_ack():			# if ack go on
                return(True)			# done
        return(False)				# all other is fail


def target_write_block(address, block):		# write a block into target memory
    l = len(block) - 1 & 0xFF			# length of block
    block = bytes([l]) + block			# leads the block bytes
    c = 0x00					# crc
    for i in range (0, l+2):			# over length and block bytes
        c = c ^ block[i] & 0xFF			#
    block = block + bytes([c])			# is appended to block as byte value
    ser.write(b'\x31')				# send the cmd 0x31
    ser.write(b'\xCE')				# end its crc
    if target_ack():				# if ack
        ser.write(target_addrblk(address))	# send the address
        if target_ack():			# if ack
            ser.write(block)            	# send the data l+data+crc
            if target_ack():			# if ack
                return(True)			# return with success
    return(False)				# otherwise return without success


def target_write(addr, blen):			# write from file into target at addr   
    data = []					# empty block
    try:					# try to open the file
        f = open(target_file, 'rb')		# open for reading bytes
        data = f.read()				# reading
        f.close()				# close file
    except:					# if not read
        print(' could not read: ', target_file) # print a message
        return(False)				# and leave with no success
    print( target_file, '(', len(data), 'bytes)') # go on

    print('  write :', end='')			# verifying starts here    
    bytecnt = len(data)				# bytes to write
    byteaddr = addr				# address to write to
    bptr = 0					# pointer into block
    bcnt = 0					# count of data in block
    while bytecnt != 0 :			# any byte to write 
        if bytecnt >= blen:			# more bytes than a block length
            bcnt = blen				# amount of bytes to write
        else:					# less bytes than a block length
            bcnt = bytecnt			# zake the rest
        pling(write_char)			# give a sign of life per block
        if not target_write_block(byteaddr,(data[bptr:(bptr+bcnt)])):
            return(False)			# return with no success if write failed
        byteaddr = byteaddr + bcnt		# update to next address
        bytecnt = bytecnt - bcnt		# update bytes to write
        bptr = bptr+bcnt			# update pointer into block to write

        
    print('\n  verify:', end='')		# verifying starts here
    bytecnt = len(data)				# bytes to verify
    byteaddr = addr                             # start address
    bptr = 0					# pointer into data
    bcnt = 0                                    # count of data in block
    while bytecnt != 0:				# as long as bytes to verify
        if bytecnt >= blen:			# more bytes than a block length
            bcnt = blen				# amount of bytes to write
        else:					# less bytes than a block length
            bcnt = bytecnt			# zake the rest        
        pling(verify_char)			# give a sign of life per block
        block = target_read_block(byteaddr, blen) # read a block from target	
        if block[:bcnt] != data[bptr:(bptr+bcnt)]: # if not equal
            return(False)			# leave without succes
        byteaddr = byteaddr + bcnt		# update to next address
        bytecnt = bytecnt - bcnt		# update bytes to verify
        bptr = bptr+bcnt			# update pointer into block to verify        

    return(True)				# return with success


def target_go(addr):				# jump to address
    ser.write(b'\x21')				# write cmd 0x21
    ser.write(b'\xDE')				# and crc
    if target_ack():				# wait for ack
        ser.write(target_addrblk(addr)) 	# write address block with crc
        if target_ack():			# if ack go on
            return(True)			# return with success
    return(False)				# all others return w/o success
    

def target_load():
    global target_file
    clearoutput()
    print('» please choose a file to flash ')
    try:					# try to get a filename
        file = askopenfile(parent=editor,title='select binary file to flash')
        target_file = file.name			# remember filename
    except:					# if not got a filename
        pass
        print(' sorry no file, please try again ')
        return()
    if target_file == "":
        print(' sorry no file, please try again ')
        return()

    print('  using:', target_file )

    uart_init_loader()                          # init UART for bootloader
    print(' ')
    print('» needed connections to program the target: ')
    print('  adapter | target    ')
    print('  --------+-----------')
    print('  3V3-GND | GND       ')
    print('  3V3-VDD | VDD       ')
    print('  3V3-TxD | RxD (PA10)')
    print('  3V3-RxD | TxD (PA09)')
    print('  3V3-DTR | /RESET (optional for automatic reset to bootmode)')
    print('  3V3-RTS | BOOT0  (optional for automatic reset to bootmode)')
    print(' ')
    print('» sequence for manual reset into bootmode without automatic:')
    print('  1. press RESET ')
    print('  2. press BOOT0 ')
    print('  3. release RESET ')
    print('  4. release BOOT0 ')
    print(' ')
    try: i = input('» and then hit [ENTER] to continue ... \n\n')
    except: pass
 
    uart_empty()                                # important: between empty and 1st read has
                                                # to be time < timeout, otherwise the UART
                                                # will be reinitialized and DTR/RTS toggled
    print('» target init: ', end = '')
    if not target_reset():
        print('no.')
        print(' please fix it and try again ')
        return()
    print('ok.')

    print('» bootloader version read: ', end = '')
    if not target_get():
        print('no.')
        print(' please fix it and try again ')
        return()
    print('ok.')

    print('» copy target memory to:', target_copy , end = '')
    if not target_read( target_addr , bcnt, blen ):
        print(' no.')
        print(' please fix it and try again ')
    print(' ok.')

    print('» global erase target: ', end = '')
    if not target_global_erase():
        print('no.')
        print(' please fix it and try again ')
        return()
    print('ok.')

    print('» flash target: ', end = '')
    if not target_write( target_addr, blen):
        print(' no.')
        print(' please fix it and try again ')
        return()
    print(' ok.')

    print('» jump to new code: ', end='')
    if not target_go(target_addr):
        print('no.')
        print(' please fix it and try again ')
    print('ok')
    print(' ')
    print('hint: on Mecrisp this jump only works if BOOT0 is released,')
    print('      because Mecrisp goes through a real reset on start. ')

# flash functions
# »-------------------------------------------------------------------«
# help functions

htext = []					# line tupel
hline = 1					# cursor line position

def help_line():				# read linenumber
    global hline
    position = (helptext.index(INSERT))		# position at cursor
    row, col = position.split('.')		# extract row number
    text = helptext.get((row + '.0'), (row + '.4')) # read linenumber
    lnr = int(text)				# readout starts at 1
    hline = lnr					#
    return(lnr-1)                               # return index 0..


def help_level(i) :				# read level from line i
    global htext				# help text is global
    line = htext[i]				# take a line
    level = int( line[1:2] )			# 2nd char in line as integer
    if level == 0 : level = 10                  # lowest level
    return(level)				# rturn the level as integer


def help_mark(i) :				# read mark from line as TRUE/FALSE
    global htext				# help text is global
    line = htext[i]				# take a line
    mark = False				# default no display
    if line[0:1] == ' ' : mark = True		# space means display
    return(mark)				# return True/False


def help_onoff(i,on) :				# set/reset mark to display
    global htext				# help text is global
    line = htext[i]				# take a line
    if on :
        htext[i] = ' ' + line[1:]		# on
    else :
        htext[i] = '#' + line[1:]		# off
    return


def help_display():				# load display area
    helptext.delete(0.0, END)			# clear display area
    pos = 0
    outcount = 0                                #
    for n in range (0 , len(htext)):		# run line by line
        line = htext[n]                         # take a line (0..n)
        if line[0:1] == ' ' :			# check for display marker

            if int( line[1:2] ) > 3 : tg = 'l0' # tag as function from
            else : tg = 'l' + line[1:2] 	# level

            line = line[2:]                     # leave out marker & level

            if line[0:3] == '** ' : line = ' · ' + line[3:]
            if line[0:3] == '***' : line = ' ··' + line[3:]  

            n = n+1                             # output (1..n+1)
            num = str(n)                        # make a number string
            if n < 1000 : num = ' ' + num       # add leading spaces
            if n < 100 : num = ' ' + num        # 
            if n < 10 : num = ' ' + num         #
            hint = ''                           # normaly no hint
            if n < len(htext) :                 # avoid range overflow
                if not help_mark(n) :		# get the mark from next line 
                    hint = ' »'                 # make a dot if hidden
            line = num + ' ' + line[:-1] + hint + '\n' # line, content        
            helptext.insert((str(n)+'.0'), line, tg) # display only with marker
            outcount = outcount + 1             #
            if hline == n : pos = outcount	#
  
    helptext.delete('end - 1 chars')		# there was an issue with last char
    helptext.mark_set("insert", "%d.%d" % (pos, 1)) # restore cursor position 
    return()					#


def help_read():				# read file and put a mark at each line
    global filestate				# filestate is global
    global htext				# the textbuffer is global
    filename = helpfile				# load the file
    try:					# try it
        file = open(filename, 'r')		# open file
        htext = file.readlines()		#
        file.close()				# close the file
        header = True                           # to show level all before first 1
        for i in range (0, len(htext)) :	# modify line by line
            line = htext[i]			# take a line
            level = 0				# define the line level
            while line[level] == '*' :		# by counting the leading stars
                level = level + 1               # count
            level = str(level)                  # make a level string
            if level == '1' : 			# default level 1 is on
                mark = ' '                      # marker for on
                header = False                  # switch off header
            else :				# default level <> 1 is off
                if header : mark = ' '          # only header space is on
                else : mark = '#'               # all others are off
            htext[i] = mark + level + line      # finaly compose the line
    except:					# try did not work
        filestate = "no help text found"	# give an error message
    return()					#

def help_close(n) :
    l = help_level(n)
    for i in range(n+1, len(htext)):
        ll = help_level(i) 
        if ll <= l : break
        else : help_onoff(i,False)

def help_open(n) :
    l = help_level(n)
    lth = 10
    for i in range(n+1, len(htext)):
        ll = help_level(i)
        if ll <= l : break
        if ll <= lth :
            help_onoff(i,True)
            lth = ll
    return()

def help_open_reverse(n) :
    l = help_level(n)
    for i in range (0, (n+1)):
        ll = help_level(n-i)
        if ll <= l :
           l = ll
           help_onoff(n-i,True)

def help_search(word):
    for n in range (0, len(htext)):
        t = htext[n]
        t = t.lower()
        word = word.lower()
        if word in t :
            help_open_reverse(n)
            help_display()
            
## events for editor actions

def patch_delete():			# patch: delete space after space
    pos1 = (helptext.index(INSERT))	# position is inserted
    row, col = pos1.split('.')		# split into row and column
    pos2 = row + '.' + str(int(col)-1)  # one step back
    helptext.delete(pos2, pos1)		# delete
    return()

def open_fold_event(event):
    help_open(help_line())
    help_display()

def close_fold_event(event):
    help_close(help_line())
    help_display()

def close_window_event(event):
    #help.destroy()
    help.withdraw() # help.iconify()

## build a help window with tkinter

help = Tk()

help.title('FEE Helper')
help.minsize(width=px_min, height=py_min)
help.maxsize(width=px_max, height=py_max)
help.geometry("1000x600+550+150")
help.config()

helptext = Text(help, takefocus=1, font=(fontname,fontsize), bg=backcolor, fg=fontcolor, insertbackground=cursorcolor, state="normal", width=tx, height=ty, blockcursor=True, insertwidth=cursorwidth, insertontime=cursor_ton, insertofftime=cursor_toff )
helptext.pack(side=LEFT,fill=BOTH, expand=1)

helpscroll = Scrollbar(helptext, orient=VERTICAL)
helpscroll.config(command = helptext.yview)
helpscroll.pack(side=RIGHT, fill=Y)

helptext.config(yscrollcommand = helpscroll.set)

helptext.tag_config('l0', foreground = '#d5d2c8')
helptext.tag_config('l1', foreground = 'orange2')
helptext.tag_config('l2', foreground = 'yellow green')
helptext.tag_config('l3', foreground = 'sky blue')

help_read()
help_display()
help.withdraw() # help.iconify()

# ---- hot-keys ----

help.bind("<Right>", open_fold_event)
help.bind("<Left>", close_fold_event)
help.bind("<F1>", close_window_event)

# help functions
# »-------------------------------------------------------------------«
# main program with reading in the loop


if os.name == 'posix':
    def KeyHit():
        return( select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], []) )

    def KeyInput():
        return( sys.stdin.read(1) )

    old_settings = termios.tcgetattr(sys.stdin)

    tty.setcbreak(sys.stdin.fileno())
    sys.stdin.flush()

else:
    def KeyHit():
        return( msvcrt.kbhit() )

    def KeyInput():
        return( msvcrt.getch() )

try:
    clearoutput()                                 # clear terminal window
    print("» exit application by closing the editor, please «") 
    nochmal = True
    while nochmal:                                # looping until nochmal=False
        read_target()
        if KeyHit():                              # act on pressed key
            c = KeyInput()                        # read it
            write_target(c)                       # and pass it to the target
        try:
            Status_update()                       # update info for status line
            editor.update()                       # update tkinter object
            crash_text = editortext.get(0.0, END) # copy in case tkinter object is closed
        except:
            break
        try: help.update()
        except: pass
        test_timer()                              # telnet timeout prevention
        time.sleep(20/1000)                       # with 50Hz

finally:

    save_copy(crash_text)                         # save last editor content

    if os.name == 'posix':
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)

    if uart == True:                              # close UART
        ser.close()
    else:                                         # or close telnet connection
        tn23.close()

    if logenable:                                 # close logfile
        close_log()

    print('\n» ¡Hasta la proxima! « \n')          # see you next time

# »-------------------------------------------------------------------«
# This is the end, there is nothing to come.